/*
 * d10.c
 * (c) 1989 by Stephen A. Ness.
 * Roland D-10 MIDI system exlusive routines.
 */

#include <stdio.h>
#include <bios.h>
#include <osbind.h>
#include <xbios.h>
#include "midi.h"
#include "d10.h"
#include "midimsg.h"
#include "roland.h"
#include "timer.h"

#if	ROLAND_D10

#if	DEBUG
#undef	Midiws
void
Midiws (n, buf) int n; unsigned char *buf;
{
	while (n-- >= 0)
		printf("%02x ", *buf++);
	printf("\n");
}
#endif

unsigned char SysEx_buf[5 + 3 + NSYSEX + 1 + 1] =	/* hdr+addr+data+csum+EOX */
	{ SysEx, M_ID_ROLAND, DEV_ID_MC, MOD_ID_D10 };

/*
 * Build a System Exclusive command in SysEx_buf.
 * This copies 'n' data bytes from 'buf' and appends csum and EOX;
 * the devid, cmd, and addr[012] bytes should have been set already.
 * Return the command length, from SysEx to EOX.
 */
int
d10_buf(buf, n) register unsigned char *buf; register int n;
{
	register unsigned char *bp;
	register int csum;

	if (n > NSYSEX)
		return error("%d System Exclusive data bytes exceeds max %d",
			n, NSYSEX);
	csum = SysEx_addr0 + SysEx_addr1 + SysEx_addr2;
	for (bp = &SysEx_data; n-- > 0; )
		csum += (*bp++ = *buf++);
	*bp++ = (-csum) & 0x7F;
	*bp++ = EOX;
	return bp - &SysEx_buf[0];
}

/*
 * Machine-specific D-10 initialization.
 */
void
d10_init()
{
	register int state, c;

	/* Write a logo on the D-10 display window and pause. */
	timerA_enab();
	d10_msg(LOGO1, LOGO2);
	pause(2);

	/*
	 * Read Note_on/Note_off sequence to check if AllNotesOff transmitted.
	 * Give up if no response in 10 seconds.
	 */
	if (vflag) {
		d10_msg("Press any key.", NULL);
		printf("Press any key on the D-10.\n");
			for (state = 1, timerA = 0; state > 0 && timerA < 10 * NTICKS; ) {
			if (Bconstat(BC_MID) != 0) {
				c = Bconin(BC_MID) & 0xFF;
				if (state == 1 && Msg(c) == NoteOn) {
					timerA = 9 * NTICKS;
					state = 2;
				} else if (state == 2 && Msg(c) == CtrlChange)
					state = 3;
				else if (state == 3 && c == AllNotesOff)
					state = -1;
				else if (state == 3)
					state = 2;
			}
		}
		if (state == -1) {
			d10_msg("Set MIDI Tx", "AllNotesOff=OFF");
			printf("Set MIDI Tx AllNotesOff=OFF\n");
			pause(5);
		}
		d10_msg(LOGO1, LOGO2);
		timerA_disab();
	}
}

/*
 * Write a two-line ASCII message to the D-10 display window.
 * Each message line can be up to 16 ASCII characters; extras are ignored.
 * Shorter lines are centered, padded with spaces to 16 characters.
 * Either argument can be NULL.
 * Return the length of the System Exclusive message.
 */
int
d10_msg(s1, s2) register char *s1, *s2;
{
	register unsigned char *cp;
	register int n;
	unsigned char buf[32];

	SysEx_devid = DEV_ID_MC;
	SysEx_cmd = DT1;
	SysEx_addr0 = 0x20;
	SysEx_addr1 = 0;
	SysEx_addr2 = 0;
	cp = &buf[0];
	if (s1 != NULL) {
		for (n = (16 - strlen(s1)) / 2; n-- > 0; )
			*cp++ = ' ';
		while (cp < &buf[16] && *s1 != '\0')
			*cp++ = (*s1++) & 0x7F;
		if (*s1 != '\0')
			fprintf(stderr, "d10_msg: line1 message truncated\n");
	}
	while (cp < &buf[16])
		*cp++ = ' ';			/* line1 space padding */
	if (s2 != NULL) {
		for (n = (16 - strlen(s2)) / 2; n-- > 0; )
			*cp++ = ' ';
		while (cp < &buf[32] && *s2 != '\0')
			*cp++ = (*s2++) & 0x7F;
		if (*s2 != '\0')
			fprintf(stderr, "d10_msg: line2 message truncated\n");
	}
	while (cp < &buf[32])
		*cp++ = ' ';			/* line2 space padding */
	n = d10_buf(buf, 32);
	Midiws(n-1, &SysEx_buf[0]);
	return n;	
}

/*
 * Read a D-10 specific voice specification.
 * The voice specification is of the form [ABabir]nn[[+-]kk].
 * Store commands to perform the voice change in a buffer,
 * return a pointer to it and store its length through np.
 * Return NULL for unrecognized specifications
 * (including MIDI program change number nn, handled by the caller).
 * The [+-]kk is an optional key shift parameter.
 * For simple [ABab] specifications, do a ProgChange.
 * For voices with [ir] specifications or key shifts,
 * do a DT1 command to the timbre temp area.
 */
unsigned char *
d10_voice(np, chan, s) int *np, chan; register char *s;
{
	static unsigned char buf[0x10];
	register int voice, c;
	int key, group;
	char *cp;

	c = *s++;
	if (c!='A' && c!='B' && c!='a' && c!='b' && c!='i' && c!='r')
		return NULL;
	voice = atoi(s) - 1;
	if (((c=='a' || c=='b' || c=='i' || c=='r') && (voice<0 || voice>63))
	 || ((c=='A' || c=='B') && (voice<11-1 || voice>88-1))) {
		error("illegal voice number");
		return NULL;
	}
	if ((cp = strchr(s, '+')) != NULL || (cp = strchr(s, '-')) != NULL) {
		key = atoi(cp);
		if (key < -24 || key > 24) {
			error("illegal key shift");
			return NULL;
		}
	} else
		key = 0;
	if (c=='A' || c=='B' || c=='a' || c=='b') {
		group = (c == 'A' || c == 'a') ? 0 : 1;
		if (c == 'A' || c == 'B')
			voice = ((voice - 10) / 10) * 8 + (voice % 10);
		if (key == 0) {
			if (c == 'B' || c == 'b')
				voice += 64;
			buf[0] = ProgChange | chan;
			buf[1] = voice;
			*np = 2;
			return &buf[0];
		}
	} else
		group = (c == 'i') ? 2 : 3;
	if (chan > 9-1) {
		error("illegal channel, must be 1 to 9 for DT1");
		return NULL;
	}
	SysEx_cmd = DT1;
	SysEx_addr0 = 0x03;
	SysEx_addr1 = (chan == 9-1) ? 0x01 : 0x00;
	SysEx_addr2 = 0x7F & (0x10 * chan);
	buf[0] = group;				/* tone group */
	buf[1] = voice;				/* tone number */
	buf[2] = 24 + key;			/* key shift */
	buf[3] = 50;				/* fine tune */
	buf[4] = 2;				/* bender range */
	buf[5] = 0;				/* assign mode */
	buf[6] = (c == 'r') ? 0 : 1;		/* reverb, off for rhythm tones */
	buf[8] = 100;				/* level */
	buf[9] = 7;				/* panpot */
	*np = d10_buf(buf, sizeof(buf));
	return &SysEx_buf[0];
}

#endif

/* end of d10.c */
