/*
 * midi.c
 * 10/6/89
 * (c) 1989 by Stephen A. Ness.
 * Usage: midi [ -msize ] [ file ... ]
 * Interactive Atari ST MIDI sequencer.
 * Load and play file ... if specified.
 * Otherwise, accept interactive input from the console;
 * see 'h' help message for interactive command descriptions.
 * Option: -msize	Set capture buffer size (default: 64K).
 *
 * This program captures and plays through the Atari ST MIDI port.
 * MIDI input is an asychronous serial data stream.
 * This program captures a byte stream in Track 0 as an array of longs,
 * each containing a 3-byte timing tick count and a MIDI input byte.
 * After capture, it translates the MIDI input into an array of 6-byte
 * entries, each containing a 3-byte tick count and a 1 to 3 byte command,
 * sorted by tick count.
 *
 * Compile:
 *	cc -o midi midi.o d10.o midifile.o midihack.o midimsg.o timer.o timerint.o
 * d10.c contains Roland D-10 specific routines.
 * midifile.c contains file input/output routines.
 * midihack.c contains special purpose hacks, kept separate for no good reason.
 * midimsg.c contains routines to dump MIDI information in semireadable form.
 * timer.c and timerint.s contain the timerA enable/disable/handling routines.
 *
 * Unused interactive commands: ikyz.
 */

#include <osbind.h>
#include <bios.h>
#include <stdio.h>
#include <ctype.h>
#if	__STDC__
#include <stddef.h>		/* should include other ANSI headers too... */
#endif
#include "midi.h"
#include "midimsg.h"
#include "timer.h"
#if	ROLAND_D10
#include "d10.h"
#endif

/* Globals. */
int		capture = 1;		/* MIDI input capture track	*/
CHANNEL		chanmap[16] = {		/* MIDI channel bitmap		*/
	0x0001, 0x0002, 0x0004, 0x0008, 0x0010, 0x0020, 0x0040, 0x0080,
	0x0100, 0x0200, 0x0400, 0x0800, 0x1000, 0x2000, 0x4000, 0x8000 };
char		cmdbuf[BUFLEN];		/* interactive command buffer	*/
EVENT		event[NTRACKS];		/* playback event list		*/
int		fflag = 1;		/* filter TimingClock messages	*/
char *		filename;		/* save/load default filename	*/
int		iargc;			/* interactive arg count	*/
char *		iargv[NARGS];		/* interactive command args	*/
unsigned long	memsize = 65536L;	/* memory size of capture buf	*/
unsigned int	mmbeats = 4;		/* metronome beats per measure	*/
unsigned int	mmmeasures = 2;		/* measures before measure 1	*/
int		mmticks;		/* ticks per metronome beat	*/
TIME		now;			/* current time			*/
CHANNEL		pchannels;		/* playback channels		*/
int		ptracks;		/* playback track count		*/
char *		tempfile;		/* temp file name		*/
int		tflag = 1;		/* timer flag			*/
FILE *		tfp;			/* temp FILE *			*/
TRACK		track[NTRACKS+1];	/* tape tracks			*/
int		vflag = 1;		/* verbose flag			*/

main(argc, argv) int argc; char *argv[];
{
	register int qflag, i;

	/* Process command line arguments. */
	if (argc > 1 && argv[1][0] == '-') {
		if (argv[1][1] != 'm') {
			fprintf(stderr, "Usage: midi [ -msize ] [ file ... ]\n");
			exit(1);
		}
		memsize = atol(&argv[1][2]);
		--argc;
		++argv;
	}

	/* Load and play specified files. */
	if (argc > 1) {
		vflag = capture = 0;
		initialize();
		for (i = 1; i < argc; i++) {
			iargc = 2;
			iargv[1] = argv[i];
			if (load(0)) {
				iargc = 1;
				begin();
				erase();
			}
			if (Bconstat(BC_CON) != 0) {
				Bconin(BC_CON);
				break;
			}
		}
		qflag = 1;
	} else {
		qflag = 0;
		tempfile = tempnam(NULL, "m");
		if ((tfp = fopen(tempfile, "w+b")) == NULL)
			fatal("cannot open temp file \"%s\"", tempfile);
		initialize();
	}

	/* Interactive input loop. */
	while (!qflag) {
		printf(PROMPT);
		fflush(stdout);
		if (fgets(cmdbuf, BUFLEN, stdin) == NULL)
			break;
		if (cmdbuf[0] == '!') {
			system(&cmdbuf[1]);
			continue;
		} else if (parse() == 0)
			continue;
		switch(*iargv[0]) {

		case '?':
			if (mustbe(0, 0))
				status(1);
			break;

		case 'a':
			if (mustbe(1, 2))
				alter();
			break;

		case 'b':
			if (mustbe(0, 2))
				begin();
			break;

		case 'c':
			if (iargc == 1) {
				capture = 0;
				puts("Capture is off.");
			} else if (iargc == 2)
				captrack(1);
			else if (mustbe(2, 3))
				control();
			break;

		case 'd':
			if (mustbe(1, 3))
				dump();
			break;

		case 'e':
			if (mustbe(0, 3))
				erase();
			break;

		case 'f':
			if (mustbe(0, 1)) {
				fflag = (iargc == 1) ? !fflag : atoi(iargv[1]);
				printf("Timing clock filter is %s.\n", (fflag) ? "on" : "off");
			}
			break;

		case 'g':
			if (mustbe(2, 3))
				getvoice();
			break;

		case 'h':
			if (!mustbe(0, 0))
				break;
			puts("\t!cmd\t\tPass cmd to system");
			puts("\t?\t\tPrint track status information");
			puts("\ta [ t ] x\tAlter timing of track t by factor x");
			puts("\tb [ x [ y ] ]\tBegin MIDI input/output at time x to time y");
			puts("\tc [ t ]\t\tCapture on track t");
			puts("\tc x y [ t ]\tControl change channel x value y to track t");
			puts("\td t [ x [ y ] ]\tDump track t from x to y");
			puts("\te [t [x [y]]]\t\Erase track t from x to y");
			puts("\tf [ x ]\t\tTiming clock filter x");
			puts("\tg x y [ t ]\tGet voice x on channel y to track t");
			puts("\th\t\tHelp information");
			puts("\tj t1 t2 t\tJoin tracks t1 and t2 to track t");
			puts("\tl [ t ] file\tLoad track t from file");
			puts("\tm [x [y [z]]]\tMetronome MM=x, y beats/measure");
			puts("\tn t x\t\tNumber track t as channel x");
			puts("\to [ t ]\t\tOff track t");
			puts("\tp [ t ]\t\tPlayback track t");
			puts("\tq\t\tQuit");
			puts("\tq t [y [z]]\tQuantize track t from y to z");
			puts("\tr x [ t [ q ] ]\tRecord from x to track t, quantize q");
			puts("\ts [ t ] dest\tSave track t to dest (track or file)");
			puts("\tt x\t\tTimer state x");
			puts("\tu file\t\tUnlink file");
			puts("\tv x\t\tVerbose flag x");
#if	ROLAND_D10

			puts("\tw l1 [ l2 [t]]\tWindow message l1 l2 to track t");
#endif
			puts("\tx\t\tAllNotesOff");
			break;

		case 'j':
			if (mustbe(3, 3))
				join();
			break;

		case 'l':
			if (mustbe(0, 2))
				load(1);
			break;

		case 'm':
			if (mustbe(0, 3))
				metronome();
			break;

		case 'n':
			if (mustbe(2, 2))
				number();
			break;
				
		case 'o':
			if (mustbe(0, 1))
				playback(0);
			break;

		case 'p':
			if (mustbe(0, 1))
				playback(1);
			break;

		case 'q':
			if (iargc == 1)
				++qflag;
			else if (mustbe(1, 3))
				quantize();
			break;

		case 'r':
			if (mustbe(1, 3))
				record();
			break;

		case 's':
			if (mustbe(0, 2))
				save();
			break;

		case 't':
			if (mustbe(0, 1)) {
				tflag = (iargc == 1) ? !tflag : atoi(iargv[1]);
				printf("Timer is %s.\n", (tflag) ? "on" : "off");
			}
			break;

		case 'u':
			if (mustbe(1, 1) && unlink(iargv[1]) == -1)
				error("cannot remove file \"%s\"", iargv[1]);
			break;

		case 'v':
			if (mustbe(0, 1)) {
				vflag = (iargc == 1) ? !vflag : atoi(iargv[1]);
				if (vflag) {
					fprintf(stderr, "midi: %s\n", VERSION);
					fprintf(stderr, "(c) 1989 by Stephen A. Ness.\n");
				}
				printf("Verbose is %s.\n", (vflag) ? "on" : "off");
			}
			break;

#if	ROLAND_D10
		case 'w':
			if (mustbe(1, 3))
				window();
			break;
#endif

		case 'x':
			if (iargc == 1)
				allnotesoff(ALLCHANS);
			else if (mustbe(2, 2))
				hack();
			break;

		default:
			error("command '%c' unrecognized\n", *iargv[0]);
			break;
		}
	}
	if (argc == 1 && unlink(tempfile) == -1)
		error("cannot remove temporary file \"%s\"", tempfile);
	exit(0);
}

/*
 * Send AllNotesOff on given channels.
 */
void
allnotesoff(chan) register CHANNEL chan;
{
	register int i;
	static unsigned char buf[3] = { CtrlChange, AllNotesOff, 0 };

	if (chan != 0)
		for (i = 0; i < 16; i++)
			if (chan & chanmap[i]) {
				buf[0] = CtrlChange|i;
				Midiws(sizeof(buf)-1, &buf[0]);
			}
}

/*
 * Alter track timing.
 * Interactive usage: a [ track ] factor.
 * The timing factor may be of the forms "x[/y]", "+time" or "-time".
 */
int
alter()
{
	register unsigned char *bp, *bpend;
	register TIME t;
	register int flag;
	TRACK *tp, *btp, *etp;
	char *s;
	unsigned int x, y;
	TIME t0;

	if (iargc == 2) {
		btp = &track[1];
		etp = &track[NTRACKS];
	} else {
		if ((tp = trackid(1, 1)) == NULL)
			return;
		btp = etp = tp;
	}
	s = iargv[iargc-1];
	if (*s == '+' || *s == '-') {
		if (vflag)
			printf("Alter timing by %s.\n", s);
		flag = (*s == '+') ? 1 : 2;
		++iargv[iargc-1];
		t0 = time(iargc-1);
		if (strchr(s, ':') == NULL && strchr(s, '.') != NULL)
			t0 = ((long)t0) - ((long)(mmmeasures-1)) * mmbeats * mmticks * TICKSPERBEAT;
	} else {
		flag = 0;
		x = atoi(s);
		y = ((s = strchr(s, '/')) == NULL) ? 1 : atoi(++s);
		if (y == 0)
			return error("divisor cannot be 0");
		if (vflag)
			printf("Alter timing by %d/%d.\n", x, y);
	}
	for (tp = btp; tp <= etp; tp++) {
		if (!inuse(tp) || !timed(tp))
			continue;
		bp = tp->t_buf;
		if (flag == 2) {
			t = etime(bp);
			if (t0 > t) {
				error("cannot alter to negative time");
				continue;
			}
		}
		for (bpend = t_end(tp); bp < bpend; ) {
			t = etime(bp);
			if (flag == 0) {
				t *= x;
				if (y != 1)
					t /= y;
			} else if (flag == 1)
				t += t0;
			else
				t -= t0;
			*bp++ = t >> 16;
			*bp++ = t >> 8;
			*bp++ = t;
			bp += 3;
		}
		tp->t_time = t;
		if (vflag) {
			printf("Track %d time=", tp - &track[0]);
			printtime(t);
			printf(".\n");
		}
	}
}

/*
 * Do the work: capture MIDI input and/or play MIDI output.
 * Interactive usage: b [ start [ stop ] ].
 */
int
begin()
{
	register unsigned char *bp, *ebp;
	register EVENT *ep;
	register int c;
	register TIME next, mmnext;
	unsigned char *bpend;
	TIME start, stop, t;
	EVENT e, *epend;
	TRACK *tp;

	start = (iargc == 1) ? (TIME)0L : time(1);
	stop = (iargc <= 2) ? (TIME)-1L : time(2);

	/* Initialize. */
	if (capture != 0) {
		if (play(&track[capture]))
			return error("cannot capture to play track");
		else if (inuse(&track[capture]))
			puts("warning: capture track in use\n");
	}
	event_setup(start, stop);
	if (ptracks) {
		if (!tflag)
			return error("cannot play with timer off");
		ep = &event[0];
		epend = &event[ptracks - 1];
		next = ep->e_time;
	} else if (!capture && !mmticks)
		return error("capture off, no playback, no metronome: no work");
	status(0);
	midi_flush();

	/* Playback and/or capture until character typed. */
	if (vflag)
		puts("Type any character to quit.");
	tp = &track[0];
	bp = tp->t_buf;
	bpend = bp + memsize;
	if (mmticks) {
		c = mmticks * TICKSPERBEAT * mmbeats;	/* ticks per measure */
		Bconout(BC_MID, (start == 0) ? Start : Continue);
		mmnext = ((start + c - 1) / c) * c;	/* next metronome tick */
	}
	if (tflag || mmticks) {
		timerA_enab();
		timerA = start;
	}
	while (Bconstat(BC_CON) == 0) {

		/* Capture. */
		if (capture && Bconstat(BC_MID) != 0) {	/* MIDI character ready */
			c = Bconin(BC_MID) & 0xFF;
			if (c == ActiveSensing)
				continue;	/* ignore active sensing */
			else if (c == TimingClock && fflag)
				continue;	/* ignore TimingClock if fflag */
			if (tflag) {
				*(long *)bp = (timerA << 8) | c;
				bp += sizeof(long);
			} else
				*bp++ = c;
			if (bp >= bpend) {
				error("output buffer full");
				break;
			}
		}

		/* Playback. */
		if (ptracks && next <= timerA) {
			ebp = ep->e_bp + 3;
			c = *ebp++;
			Bconout(BC_MID, c);
			if (Msg(c) < ProgChange || Msg(c) == PitchBend || c == SongPosPtr) {
				Bconout(BC_MID, *ebp++);
				Bconout(BC_MID, *ebp);
			} else if (Msg(c) == ProgChange || Msg(c) == ChanPr || c == SongSelect)
				Bconout(BC_MID, *ebp);
			ebp = (ep->e_bp += 6);
			if (ebp < ep->e_end) {
				next = ep->e_time = etime(ebp);
				if (ep < epend && next > (ep+1)->e_time) {
					for (e = *ep; ep < epend && next > (ep+1)->e_time; ) {
						*ep = *(ep+1);
						ep++;
					}
					*ep = e;
				}
			} else {
				while (ep < epend) {
					*ep = *(ep+1);
					ep++;
				}
				--epend;
				if (--ptracks == 0 && !capture)
					break;
			}
			ep = &event[0];
			next = ep->e_time;
		}
			
		/* Timing clocks. */
		if (mmticks && timerA >= mmnext) {
			mmnext += mmticks;
			Bconout(BC_MID, TimingClock);
		}

	}
	if (tflag || mmticks) {
		now = timerA;
		timerA_disab();
		if (mmticks)
			Bconout(BC_MID, Stop);
	}
	if (Bconstat(BC_CON) != 0)
		Bconin(BC_CON);
	allnotesoff(pchannels);
	if (vflag) {
		printf("Time ");
		printtime(now);
		printf(".%03ld", now % NTICKS);
		if (mmticks) {
			t = now/(mmticks*TICKSPERBEAT);		/* beats */
			if (t >= mmmeasures * mmbeats) {
				t -= mmmeasures * mmbeats;	/* without leadin */
				printf(", measure %ld.%ld.",
					(t / mmbeats) + 1, (t % mmbeats) + 1);
			}
		}
		putchar('\n');
	}
	if (capture) {
		tp->t_size = (unsigned long)bp - (unsigned long)(tp->t_buf);
		if (vflag)
			printf("Input size=%ld.\n", tp->t_size);
		if (tp->t_size > 0)
			transform();
	}
}

/*
 * Turn off capture if given track is capture track.
 */
void
capcheck(tp) register TRACK *tp;
{
	if (tp != &track[capture])
		return;
	capture = 0;
	puts("Capture is off.");
}

/*
 * Set capture on track iargv[i].
 * Return capture track number on success, 0 on failure.
 */
int
captrack(i) register int i;
{
	i = atoi(iargv[i]);
	if (i < 1 || i > NTRACKS)
		return error("capture track must be 1 to %d", NTRACKS);
	else if (play(&track[i]))
		return error("cannot capture to play track");
	printf("Capture on track %d.\n", capture = i);
	return i;
}

/*
 * Perform control change.
 * Interactive usage: c chan val [ track ].
 * The val must be of the form [pve]n, 0 <= n <= 127.
 */
int
control()
{
	register TRACK *tp;
	register int chan, cmd, val;
	unsigned char buf[3];

	if ((chan = atoi(iargv[1])) < 1 || chan > 16)
		return error("channel number must be between 1 and 16");
	--chan;
	cmd = *iargv[2]++;
	
#if	ROLAND_D10
	if (cmd == 'p' && (strchr(iargv[2], '+') != NULL || strchr(iargv[2], '-') != NULL)) {
		val = (atoi(iargv[2]) + 7) * 9;
		if (val < 0)
			val = 0;
		else if (val > 127)
			val = 127;
	} else
#endif
	if ((val = atoi(iargv[2])) < 0 || val > 127)
		return error("value must be between 0 and 127");
	if (cmd == 'v')
		cmd = Volume;
	else if (cmd == 'p')
		cmd = Pan;
	else if (cmd == 'e')
		cmd = Expression;
	else
		return error("unrecognized command, must be [vpe]");
	buf[0] = CtrlChange|chan;
	buf[1] = cmd;
	buf[2] = val;
	Midiws(sizeof(buf)-1, &buf[0]);
	if (iargc > 3 && (tp = trackid(3, -1)) != NULL)
		untimed(tp, &buf[0], sizeof(buf), chan);
}

/*
 * Dump a track to the console in semireadable form.
 * Useful for debugging and understanding device output.
 * Interactive usage: d track [ start [ stop ] ].
 */
void
dump()
{
	register unsigned char *bp, *bpend;
	register TIME t, told;
	register TRACK *tp;
	register int tptimed;

	if ((tp = trackid(1, 1)) == NULL)
		return;
	puts("Type any character to quit.");
	if (vflag)
		printf("\nDump of track %d:\n", tp - &track[0]);
	told = 0;
	bp = (iargc > 2) ? findtime(tp, time(2)) : tp->t_buf;
	bpend = (iargc > 3) ? findtime(tp, time(3)) : t_end(tp);
	tptimed = timed(tp);
	while (Bconstat(BC_CON) == 0 &&  bp < bpend) {
		if (tptimed) {
			t = etime(bp);
			printtime(t);
			printf(".%03ld ", t % NTICKS);
			if (told == 0)
				printf("         ");
			else
				printf("%6ld   ", t-told);
			msg_print(bp+3);
			told = t;
			bp += 6;
		} else
			bp = msg_print(bp);
	}
	if (Bconstat(BC_CON) != 0)
		Bconin(BC_CON);
}

/*
 * Erase one track or all tracks.
 * Interactive usage: e track [ start [ stop ] ].
 */
int
erase()
{
	register TRACK *tp;
	register unsigned char *bp1, *bp2;
	unsigned long size;
	unsigned char *bpend;

	if (iargc == 1) {
		/* Erase all tracks. */
		for (tp = &track[1]; tp <= &track[NTRACKS]; tp++)
			erase_track(tp);
		return;
	}
	if ((tp = trackid(1, 1)) == NULL)
		return;
	else if (iargc == 2)
		return erase_track(tp);		/* erase one track */

	/* Erase one track from time1 to time2. */
	bpend = t_end(tp);
	bp1 = findtime(tp, time(2));
	bp2 = (iargc > 3) ? findtime(tp, time(3)) : bpend;
	if (bp1 >= bp2)
		return;
	size = (unsigned long)bp2 - (unsigned long)bp1;
	tp->t_size -= size;
	if (tp->t_size == 0)
		return erase_track(tp);
	else if (vflag)
		printf("Track size=%ld.\n", tp->t_size);
	if (bp2 != bpend) {
		/* Copy to and from temp area to avoid overlap. */
		size = (unsigned long)bpend - (unsigned long)bp2;
		if (size > memsize)
			fatal("erase botch, size=%ld memsize=%ld", size, memsize);
		lmemcpy((char *)(track[0].t_buf), (char *)bp2, size);
		lmemcpy((char *)bp1, (char *)(track[0].t_buf), size);
	}
	lrealloc(tp->t_buf, tp->t_size);
}

/*
 * Erase one track.
 */
int
erase_track(tp) register TRACK *tp;
{
	tp->t_chan = tp->t_flags = 0;
	tp->t_size = tp->t_time = 0L;
	if (tp->t_buf != NULL) {
		free(tp->t_buf);
		tp->t_buf = NULL;
	}
}

/*
 * Print a nonfatal error message.
 * Uses the nonportable, nonANSI "%r" format.
 */
#if	__STDC__
int
error(char *s, ...)
#else
int
error(s) char *s;
#endif
{
	fprintf(stderr, "midi: %r\n", &s);
	return 0;
}

/*
 * Compare the times of two EVENTs, for qsort.
 */
int
event_compare(ep1, ep2) EVENT *ep1, *ep2;
{
	register long t1, t2;

	t1 = ep1->e_time;
	t2 = ep2->e_time;
	return (t1 < t2) ? -1 : (t1 == t2) ? 0 : 1;
}

/*
 * Set up the event list.
 * The event list contains one entry for each playing timed track,
 * sorted by time.
 */
int
event_setup(start, stop) TIME start, stop;
{
	register unsigned char *ebp;
	register EVENT *ep;
	register TRACK *tp;

	ptracks = 0;			/* playback track count */
	pchannels = 0;			/* playback channels */
	ep = &event[0];
	for (tp = &track[1]; tp <= &track[NTRACKS]; tp++) {
		if (play(tp)) {
			pchannels |= tp->t_chan;
			if (!timed(tp)) {
				/* Play untimed tracks at time 0. */
				if (start != 0)
					continue;
				if (tp->t_size > MAX_UINT)
					error("untimed track exceeds 64K");
				else
					Midiws((unsigned)tp->t_size - 1, tp->t_buf);
				continue;
			}
			ebp = findtime(tp, start);
			ep->e_end = findtime(tp, stop);
			if (ebp >= ep->e_end)
				continue;
			++ptracks;
			ep->e_bp = ebp;
			ep->e_time = etime(ebp);
			ep++;
		}
	}
	if (ptracks > 1)
		qsort(event, ptracks, sizeof(EVENT), event_compare);
	return ptracks;
}

/*
 * Print a fatal error message.
 * Uses the nonportable, nonANSI "%r" format.
 * Turn off the timer, just in case.
 */
#if	__STDC__
void
fatal(char *s, ...)
#else
void
fatal(s) char *s;
#endif
{
	fprintf(stderr, "midi: %r\n", &s);
	timerA_disab();
	unlink(tempfile);
	exit(1);
}

/*
 * Find the first entry in a timed track at or after the given time.
 */
unsigned char *
findtime(tp, t) TRACK *tp; register TIME t;
{
	register TIME t1;
	register unsigned char *bp, *bpend;

	for (bp = tp->t_buf, bpend = t_end(tp); bp < bpend; bp += 6) {
		t1 = etime(bp);
		if (t1 >= t)
			break;
	}
	return bp;
}

/*
 * Perform a ProgChange to get a new voice.
 * Interactive usage: g voice chan [ track ].
 */
int
getvoice()
{
	register int chan, voice;
	register TRACK *tp;
	int n;
	unsigned char *bp;
	unsigned char buf[2] = { ProgChange, 0 };

	if ((chan = atoi(iargv[2])) < 1 || chan > 16)
		return error("channel number must be between 1 and 16");
	--chan;
#if	ROLAND_D10
	if ((bp = d10_voice(&n, chan, iargv[1])) == NULL) {
#endif
		voice = atoi(iargv[1]);
		if (voice < 0 || voice > 127)
			return error("voice number must be between 0 and 127");
		buf[0] = ProgChange | chan;
		buf[1] = voice;
		bp = &buf[0];
		n = 2;
#if	ROLAND_D10
	}
#endif
	Midiws(n - 1, bp);
	if (iargc > 3 && (tp = trackid(3, -1)) != NULL)
		untimed(tp, bp, n, chan);
}

/*
 * Initialization.
 */
void
initialize()
{
	track[0].t_buf = xalloc(memsize);
	track[0].t_size = memsize;
#if	ROLAND_D10
	d10_init();
#endif
	midi_flush();
}

/*
 * Join tracks.
 * Interactive usage: j oldtrack1 oldtrack2 newtrack.
 */
int
join()
{
	register unsigned char *bp0, *bp1, *bp2;
	unsigned char *bp1end, *bp2end;
	TIME t1, t2;
	TRACK *tp0, *tp1, *tp2, t;
	int flag;

	if ((tp1 = trackid(1, 1)) == NULL
	 || (tp2 = trackid(2, 1)) == NULL
	 || (tp0 = trackid(3, -1)) == NULL)
		return;
	else if (timed(tp1) != timed(tp2))
		return error("cannot join timed and untimed tracks");
	if (tp0 == tp1 || tp0 == tp2) {
		flag = 1;
		if (tp0 == tp2) {
			tp2 = tp1;	/* switch to force tp0 == tp1 */
			tp1 = tp0;
		}
		tp0 = &t;		/* join to t, copied to tp1 later */
	} else {
		flag = 0;
		if (inuse(tp0))
			return error("track in use");
	}
	tp0->t_size = tp1->t_size + tp2->t_size;
	if (vflag)
		printf("Track size %ld.\n", tp0->t_size);
	bp0 = tp0->t_buf = xalloc(tp0->t_size);
	bp1 = tp1->t_buf;
	bp2 = tp2->t_buf;
	if (!timed(tp1)) {
		/* Concatenate untimed tracks. */
		lmemcpy((char *)bp0, (char *)bp1, tp1->t_size);
		lmemcpy((char *)(bp0 + tp1->t_size), (char *)bp2, tp2->t_size);
	} else {
		/* Merge timed tracks. */
		t1 = etime(bp1);
		t2 = etime(bp2);
		bp1end = t_end(tp1);
		bp2end = t_end(tp2);
		while (bp1 < bp1end && bp2 < bp2end) {
			if (t1 <= t2) {
				memcpy(bp0, bp1, 6);
				if ((bp1 += 6) < bp1end)
					t1 = etime(bp1);
			} else {
				memcpy(bp0, bp2, 6);
				if ((bp2 += 6) < bp2end)
					t2 = etime(bp2);
			}
			bp0 += 6;
		}
		if (bp1 < bp1end)
			memcpy(bp0, bp1, bp1end - bp1);
		else if (bp2 < bp2end)
			memcpy(bp0, bp2, bp2end - bp2);
	}
	tp0->t_flags = tp1->t_flags | tp2->t_flags;
	tp0->t_chan = tp1->t_chan | tp2->t_chan;
	tp0->t_time = (tp1->t_time >= tp2->t_time) ? tp1->t_time : tp2->t_time;
	if (flag) {
		free(tp1->t_buf);
		*tp1 = *tp0;
	}
}

/*
 * Like memcpy(), but with unsigned long third argument and no return value.
 */
void
lmemcpy(p1, p2, size) register char *p1, *p2; register unsigned long size;
{
	while (size > MAX_UINT) {
		memcpy(p1, p2, MAX_UINT);
		p1 += MAX_UINT;
		p2 += MAX_UINT;
		size -= MAX_UINT;
	}
	memcpy(p1, p2, (unsigned int)size);
}

/*
 * Set the metronome.
 * Interactive usage: m [ mm [ beats [ measures ] ] ].
 * This sets an integer tick count, so roundoff error
 * means the specified time may be slightly inaccurate.
 * The granularity of the tick count means inaccuracy can be worse
 * for higher metronome speeds.
 * With a fairly fast timing tick, it does not matter much.
 */
void
metronome()
{
	register int n;

	n = (iargc == 1) ? 0 : atoi(iargv[1]);
	if (n == 0) {
		mmticks = 0;
		puts("Metronome off.");
		return;
	} else
		mmticks = 60L * NTICKS / (n * TICKSPERBEAT);
	if (iargc > 2)
		mmbeats = atoi(iargv[2]);
	if (iargc > 3)
		mmmeasures = atoi(iargv[3]);
	printf("MM=%d, %d beats per measure, %d lead-in measures.\n",
		(int)(60. * NTICKS / (TICKSPERBEAT * mmticks)), mmbeats, mmmeasures);
}

/*
 * Flush the MIDI input buffer.
 */
void
midi_flush()
{
	register int flag;

	for (flag = 0; Bconstat(BC_MID) != 0; )
		if ((Bconin(BC_MID) & 0xFF) == TimingClock)
			flag = 1;
	if (flag) {
#if	ROLAND_D10
		d10_msg("Set clock mode", "to MIDI!");
#endif
		puts("Set clock mode to MIDI!");
	}
}

/*
 * Print error message unless the arg count is correct.
 * Return 0 on failure, 1 on success.
 */
int
mustbe(min, max) register int min, max;
{
	if (iargc-1 >= min && iargc-1 <= max)
		return 1;
	if (min == max)
		return error("%d argument%s required for '%c' command",
			max, (max==1) ? "" : "s", *iargv[0]);
	else
		return error("%d to %d arguments required for '%c' command",
			min, max, *iargv[0]);
}
	
/*
 * Change the channel number of a track.
 * Interactive usage: n track chan.
 */
int
number()
{
	register unsigned char *bp, *bpend, cmd;
	register TRACK *tp;
	register int n;

	if ((tp = trackid(1, 1)) == NULL)
		return;
	else if ((n = atoi(iargv[2])) < 1 || n > 16)
		return error("channel number must be between 1 and 16");
	--n;
	for (bp = tp->t_buf, bpend = t_end(tp); bp < bpend; bp += 6) {
		cmd = bp[3];
		if (!isSystem(cmd))
			bp[3] = Msg(cmd) | n;
	}
	tp->t_chan = chanmap[n];
}

/*
 * Parse the command string in cmdbuf into whitespace separated arguments.
 * Store the arguments in global iargv[] and argument count in global iargc.
 * Return the arg count.
 * Special hack for argv[0], known to be a single character:
 * e.g. parse "p1 2 3" as "p 1 2 3".
 */
int
parse()
{
	register char *s;
	register int c, inarg;
	static char cmd[2];

	s = &cmdbuf[0];
	for (iargc = 0; iargc < NARGS; )
		iargv[iargc++] = NULL;
	for (inarg = iargc = 0; (c = *s) != '\0'; s++) {
		if (isspace(c)) {
			inarg = 0;
			*s = '\0';
			if (c == '\n')
				break;
		} else if (!inarg) {
			++inarg;
			if (iargc == NARGS) {
				error("more than %d arguments, excess ignored", NARGS);
				break;
			} else if (iargc == 0 && *(s+1) != '\0') {
				inarg = 0;
				cmd[0] = c;
				cmd[1] = '\0';
				iargv[iargc++] = &cmd[0];
			} else if (c == '"') {
				iargv[iargc++] = ++s;
				while ((c = *s) != '"' &&  c != '\0')
					++s;
				if (c == '"')
					*s = '\0';
				else {
					error("missing closing quote");
					break;
				}
			} else
				iargv[iargc++] = s;
		}
	}
	return iargc;
}

/*
 * Set or clear the playback flag for a given track.
 * Interactive usage: [ o | p ] [ track ].
 */
int
playback(flag) int flag;
{
	register TRACK *tp, *btp, *etp;

	if (iargc == 2) {
		/* Specific track. */
		if ((tp = trackid(1, 1)) == NULL)
			return;
		btp = etp = tp;
	} else {
		/* All tracks. */
		btp = &track[1];
		etp = &track[NTRACKS];
	}
	for (tp = btp; tp <= etp; tp++)
		if (inuse(tp)) {
			if (flag) {
				tp->t_flags |= T_PLAY;
				capcheck(tp);
			} else
				tp->t_flags &= ~T_PLAY;
		}
}

/*
 * Print the time.
 */
void
printtime(t) register TIME t;
{
	register TIME minutes;

	minutes = t / (60 * NTICKS);
	printf("%ld:%02ld", minutes, t / NTICKS - 60 * minutes);
}

/*
 * Quantize one track or all tracks.
 * Quantizing moves events to the nearest beat or fraction of a beat.
 * Interactive usage:
 *	q qn			Set n notes per beat
 *	q dn			Set NoteOff duration n, -100 to 100
 *	q t [ start [ stop ] ]	Quantize track t from start to stop
 *	q - [ start [ stop ] ]	Quantize all tracks from start to stop
 * NoteOff duration specifies the timing of NoteOff
 * to allow some control over legato/staccato:
 *	-50	move back 1/2 (e.g., play quarter note as eighth note)
 *	0	leave unchanged (legato) [default]
 *	50	move forward 1/2 (e.g., play quarter note as dotted quarter)
 */
int
quantize()
{
	register TRACK *tp;
	register TIME start, stop;
	static unsigned int nbp = 0;
	static int dur = 0;
	unsigned int q;
	int xflag, c;

	if (mmticks == 0)
		return error("cannot quantize with metronome off");
	xflag = 0;
	c = *iargv[1];
	if (iargc == 2) {
		if (c == 'q') {
			nbp = atoi(&iargv[1][1]);
			++xflag;
		} else if (c == 'd') {
			dur = atoi(&iargv[1][1]);
			if (dur < -100 || dur > 100) {
				error("duration %d must be -100 to 100", dur);
				dur = 0;
			}
			++xflag;
		}
	}
	if (nbp == 0)
		return error("cannot quantize %d notes per beat", nbp);
	q = mmticks * TICKSPERBEAT / nbp;
	if (q * nbp != mmticks * TICKSPERBEAT)
		return error("cannot quantize %d notes per beat", nbp);
	if (xflag)
		return;
	start = (iargc < 3) ? (TIME)0L : time(2);
	stop = (iargc < 4) ? (TIME)-1L : time(3);
	if (c == '-') {
		/* Quantize all tracks. */
		for (tp = &track[1]; tp <= &track[NTRACKS]; tp++)
			if (inuse(tp) && timed(tp))
				quant_track(tp, q, dur*q/100, start, stop);
	} else {
		/* Quantize specific track. */
		if ((tp = trackid(1, 1)) == NULL)
			return;
		else if (!timed(tp))
			return error("track not timed");
		else
			quant_track(tp, q, dur*q/100, start, stop);
	}
}		

/*
 * Quantize a track.
 * If the NoteOff shift d is nonzero, the track may need to be sorted;
 * e.g. simultaneous NoteOn... Noteoff... might become Noteoff... Noteon...
 */
int
quant_track(tp, q, d, start, stop) TRACK *tp; unsigned int q; int d; TIME start, stop;
{
	register unsigned char *bp, *bpend;
	register TIME t;
	unsigned int q2, msg;

	q2 = q / 2;
	bp = findtime(tp, start);
	bpend = findtime(tp, stop);
	while (bp < bpend) {
		t = etime(bp);
		t = ((t + q2) / q ) * q;
		if (d != 0) {
			msg = Msg(bp[3]);
			if (msg == NoteOff || (msg == NoteOn && bp[5] == 0))
				t += d;
		}
		*bp++ = t >> 16;
		*bp++ = t >> 8;
		*bp++ = t;
		bp += 3;
	}
	if (d != 0) {
		if (tp->t_size / 6 > MAX_UINT)
			fatal("track too large to sort");
		qsort(tp->t_buf, (unsigned int)(tp->t_size/6), 6, time_compare);
	}
}

/*
 * Record.
 * Interactive usage: r time [ rtrack [ quant ] ]
 * Capture starting at the given time.
 * When capture is done, quantize the capture track (if desired)
 * and join it to the record track.
 */
int
record()
{
	static int qflag;
	static int rtrack;
	TIME t, t0;

	if (capture == 0)
		return error("no capture track");
	t = time(1);
	if (iargc > 2) {
		if (trackid(2, -1) == NULL)
			return;
		rtrack = atoi(iargv[2]);
		if (capture == rtrack)
			return error("capture and record tracks must be different");
	} else if (rtrack == 0)
		return error("no record track");
	if (iargc > 3) {
		if ((qflag = atoi(iargv[3])) > 0) {
			sprintf(cmdbuf, "q q%d", qflag);
			parse();
			quantize();
		}
	}

	/* Erase the record track from the starting point. */
	if (inuse(&track[rtrack])) {
		sprintf(cmdbuf, "e %d %ld", rtrack, t);
		parse();
		erase();
	}

	/* Resume capture, one lead-in measure if appropriate. */
	t0 = (t < mmbeats * mmticks * TICKSPERBEAT) ? t
		: t - mmbeats * mmticks * TICKSPERBEAT;
	sprintf(cmdbuf, "b %ld", t0);
	parse();
	begin();

	/* Quantize the capture track if desired. */
	if (qflag > 0) {
		sprintf(cmdbuf, "q %d", capture);
		parse();
		quantize();
	}

	/* Erase the capture track to the starting point. */
	sprintf(cmdbuf, "e %d %ld %ld", capture, t0, t);
	parse();
	erase();
	if (!inuse(&track[capture]))
		return;

	/* Join the capture track to the record track. */
	if (inuse(&track[rtrack])) {
		sprintf(cmdbuf, "j %d %d %d", rtrack, capture, rtrack);
		parse();
		join();
	} else {
		sprintf(cmdbuf, "s %d %d", capture, rtrack);
		parse();
		save();
		track[rtrack].t_flags |= T_PLAY;
	}
}

/*
 * Print track status (if verbose or if flag).
 * Interactive usage: ?.
 */
void
status(flag) register int flag;
{
	register TRACK *tp;
	register int flags, i, n, hflag;

	if (!flag && !vflag)
		return;
	if (!tflag)
		puts("Timer is off.");
	if (capture)
		printf("Capture on track %d.\n", capture);
	else
		puts("Capture is off.");
	if (mmticks)
		printf("MM=%d, %d beats per measure, %d lead-in measures.\n",
			(int)(60. * NTICKS / (TICKSPERBEAT * mmticks)), mmbeats, mmmeasures);
	for (hflag = n = 1, tp = &track[1]; n <= NTRACKS; n++, tp++)
		if (n == capture || inuse(tp)) {
			if (hflag) {
				puts("Track   Mode  Channel   Size  Time");
				puts("-----   ----  -------   ----  ----");
				hflag = 0;
			}
			printf("  %d:  ", n);
			flags = tp->t_flags;
			if ((flags & T_INUSE) == 0) {
				puts("Unused                        <- Record");
				continue;
			} else if (flags & T_PLAY)
				printf("  Play");
			else
				printf("   Off");
			printf("   ");
			for (i = 0; i < 16; i++)
				if (chanmap[i] == tp->t_chan)
					break;
			if (tp->t_chan == 0)
				printf("(none)  ");
			else if (i < 16)
				printf("    %2d  ", i+1);
			else
				printf("0x%04x  ", tp->t_chan);
			printf("%5ld  ", tp->t_size);
			printtime(tp->t_time);
			if (n == capture)
				printf("  <- Record");
			putchar('\n');
#if	0
			printf("%02x   ", flags);
			printf("%lx\n", tp->t_buf);
#endif
		}
}

/*
 * Read a time argument iargv[n] and return a value in ticks.
 * The time may be in any of the following forms:
 *	.		time of previous stop
 *	nnnn		raw tick count
 *	[mm]:ss[.ttt]	minutes, seconds, ticks
 *	n.[b]		measure n beat b
 * Time beginning with ".+" or ".-" is now-relative rather than absolute.
 */
TIME
time(n) int n;
{
	register TIME l;
	register char *s, *p;
	register int rel;

	s = iargv[n];
	if (*s == '.') {
		if (*++s == '\0')
			return now;
		else if (*s == '+')
			rel = 1;
		else if (*s == '-')
			rel = -1;
		else
			return error("time form '.' requires '+' or '-'");
		++s;
	} else
		rel = 0;
	if ((p = strchr(s, ':')) != NULL) {
		/* Clock time. */
		l = (atol(s) * 60 + atol(++p)) * NTICKS;
		if ((p = strchr(s, '.')) != NULL)
			l += atol(++p);
	} else if ((p = strchr(s, '.')) != NULL) {
		/* Measured time. */
		if (mmticks == 0)
			return error("no metronome, cannot count in measures");
		l = (atol(s) - 1 + mmmeasures) * mmbeats;
		if (*++p)
			l += atol(p) - 1;
		l *= mmticks * TICKSPERBEAT;
	} else	/* Raw time. */
		l = atol(s);
	if (rel == 0)
		return l;
	else if (rel > 0)
		return now + l;
	else if (now >= l)
		return now - l;
	else
		return 0L;
}

/*
 * Compare the times of two track buffer entries.
 * Called indirectly by qsort.
 */
int
time_compare(p1, p2) register unsigned char *p1, *p2;
{
	register TIME t1, t2;

	t1 = etime(p1);
	t2 = etime(p2);
	return (t1 < t2) ? -1 : (t1 == t2) ? 0 : 1;
}

/*
 * Look for a track number in iargv[n].
 * Issue an error if flag==1 and the track is not in use,
 * or if flag==0 and the track is in use;
 * flag==-1 means any track will do.
 * Return a pointer to the TRACK or NULL on failure.
 */
TRACK *
trackid(n, flag) register int n, flag;
{
	register TRACK *tp;

	if ((n = atoi(iargv[n])) < 1 || n > NTRACKS) {
		error("unrecognized track, must be 1 to %d", NTRACKS);
		return NULL;
	}
	tp = &track[n];
	if (flag == 1 && !inuse(tp)) {
		error("track %d not in use", n);
		return NULL;
	} else if (flag == 0 && inuse(tp)) {
		error("track %d in use", n);
		return NULL;
	} else
		return tp;
}

/*
 * Transform a MIDI capture file to internal representation.
 * Move a real-time command which interrupts a multibyte command
 * to precede the multibyte command.
 * Change the time of multibyte commands to the time of the last byte.
 * Insert implicit running status bytes.
 * In capture representation, each byte is time-tagged with a three-byte
 * tick count and the result stashed in a long.
 * In internal representation, each command consists of a three-byte time
 * followed by a three-byte command; extra bytes of one-byte and two-byte
 * commands are ignored.
 */
void
transform()
{
	register unsigned long nitems;
	register unsigned char *bp, *ebp;
	unsigned char b;
	CHANNEL channel;
	TIME t;
	TRACK *tp;
	static unsigned char cmd, data[6], *dp, *dpend;

	bp = track[0].t_buf;
	tp = &track[capture];
	if (!tflag) {
		/* Untimed. */
		tp->t_flags = T_INUSE;
		tp->t_size = track[0].t_size;
		if (vflag)
			printf("Track size=%ld.\n", tp->t_size);
		/* UNDONE: channel of untimed track is 0 for now. */
		tp->t_time = tp->t_chan = 0;
		if (tp->t_buf != NULL)
			free(tp->t_buf);
		tp->t_buf = xalloc(tp->t_size);
		lmemcpy((char *)(tp->t_buf), (char *)(track[0].t_buf), tp->t_size);
		return;
	}

	/* Timed. */
	if (rewind(tfp) == EOF)
		fatal("cannot rewind temp file");
	channel = 0;
	nitems = 0;
	dp = NULL;
	for (ebp = t_end(&track[0]); bp < ebp; ) {
		data[0] = *bp++;
		data[1] = *bp++;
		data[2] = *bp++;
		data[3] = b = *bp++;
		if (isData(b)) {
			/* Data byte. */
			if (dp == NULL)
				fatal("no running status, data==%02x", b);
			*dp++ = b;
			if (dp != dpend)
				continue;	/* more data follows */
			data[3] = cmd;		/* restore command */
			dp = &data[4];		/* and pointer for later */
		} else if (!isSystem(b)) {
			/* Channel mode status byte, data to follow. */
			channel |= chanmap[Chan(b)];
			cmd = b;
			dp = &data[4];
			if (Msg(b) == ProgChange || Msg(b) == ChanPr)
				dpend = &data[5];
			else
				dpend = &data[6];
			continue;
		} else if (isSysReal(b)) {
			/* System realtime status byte. */
		} else if (isSysComm(b)) {
			/* System common status byte. */
			if (b == SongPosPtr) {
				cmd = b;
				dp = &data[4];
				dpend = &data[6];
				continue;
			} else if (b == SongSelect) {
				cmd = b;
				dp = &data[4];
				dpend = &data[5];
				continue;
			} else
				dp = NULL;
		} else if (isSysEx(b)) {
			/* System exclusive status byte. */
			fatal("system exclusive capture undone");
		} else
			fatal("transform botch, b=%x", b);
		if (fwrite(&data[0], 1, 6, tfp) != 6) {
			error("temp file write error");
			return;
		}
		++nitems;
	}
	t = etime(data);
	if (vflag) {
		printf("Track size=%ld.\n", nitems * 6);
		printf("Track time=");
		printtime(t);
		printf(".\n");
	}
	tp = &track[capture];
	tp->t_flags = T_INUSE | T_TIMED;
	tp->t_size = nitems * 6;
	tp->t_time = t;
	tp->t_chan = channel;
	if (tp->t_buf != NULL)
		free(tp->t_buf);
	tp->t_buf = xalloc(tp->t_size);
	if (rewind(tfp) == EOF)
		fatal("cannot rewind temp file");
	if (fread(tp->t_buf, 6, (unsigned int)nitems, tfp) != nitems)
		fatal("temp file read error");
}

/*
 * Append n bytes of untimed control information from buf
 * to an untimed track.
 */
int
untimed(tp, buf, n, chan) TRACK *tp; unsigned char *buf; unsigned int n, chan;
{
	register unsigned long osize;
	register unsigned char *obp;

	if (timed(tp))
		return error("cannot write control info to timed track");
	if (!inuse(tp)) {
		tp->t_flags = T_INUSE;
		tp->t_time = tp->t_chan = 0;
	}
	osize = tp->t_size;
	obp = tp->t_buf;
	tp->t_size = osize + n;
	if (chan != -1)
		tp->t_chan |= chanmap[chan];
	if (vflag)
		printf("Track size %ld.\n", tp->t_size);
	tp->t_buf = xalloc(tp->t_size);
	if (osize != 0) {
		lmemcpy((char *)(tp->t_buf), (char *)obp, osize);
		free(obp);
	}
	memcpy(tp->t_buf + osize, buf, n);
}

#if	ROLAND_D10
/*
 * Put an identification message on the D-10 window.
 * Interactive usage: w line1 line2 [ track ].
 * The message goes to the window immediately,
 * and also to a specified track if three args.
 */
int
window()
{
	register TRACK *tp;
	register unsigned int n;

	n = d10_msg(iargv[1], (iargc > 2) ? iargv[2] : NULL);
	if (iargc > 3 && (tp = trackid(3, -1)) != NULL)
		untimed(tp, &SysEx_buf[0], n, -1);
}
#endif

/*
 * Allocate memory.
 * Die if allocation fails.
 */
Void *
xalloc(size) unsigned long size;
{
	register Void *p;

	if ((p = lmalloc(size)) != NULL)
		return p;
	fatal("out of memory: size=%ld memsize=%ld", size, memsize);
}

/* end of midi.c */
