/*
 * midifile.c
 * (c) 1989 by Stephen A. Ness.
 * Save and load MIDI data files.
 */

#include <stdio.h>
#include <ctype.h>
#include "midi.h"

/*
 * Single track saved file format:
 *	int magic number ST_MAGIC
 *	TRACK
 *	t.t_size data bytes
 * Multiple track saved file format:
 *	int magic number MT_MAGIC
 *	int track number
 *	TRACK
 *	t.t_size data bytes
 * with the last three items repeated for each track in use.
 * In both formats, the t_buf member of the TRACK is meaningless.
 */

/*
 * Load a saved file.
 * Interactive usage: l [ [ track ] file ].
 * Return 1 for success, 0 for failure.
 * Print a "not found" error message if flag.
 * If a single track is loaded, this does not change other tracks.
 * If all tracks are loaded, this erases all current tracks.
 */
int
load(flag) int flag;
{
	register TRACK *tp;
	register FILE *ifp;
	int i, single;
	char *name, *type;

	single = 0;
	if (iargc == 3) {
		++single;
		name = iargv[2];
	} else if (iargc == 2) {
		name = iargv[1];
		filename = xalloc((unsigned long)strlen(name) + 1);
		strcpy(filename, name);
	} else {
		if (filename == NULL)
			return error("no default filename");
		name = filename;
	}
	if ((ifp = fopen(name, "rb")) == NULL) {
		if (flag)
			error("cannot open input file \"%s\"", name);
		return 0;
	}
	if (fread(&i, sizeof(int), 1, ifp) != 1)
		return error("read error");
	if ((single && i != ST_MAGIC) || (!single && i != MT_MAGIC)) {
		type = (i == ST_MAGIC) ? "single track"
		     : (i == MT_MAGIC) ? "multitrack"
		     : "not MIDI";
		return error("%strack data file required, \"%s\" is %s data",
			(single) ? "single " : "multi", name, type);
	}
	if (vflag) {
		printf("Load %strack file \"%s\": ",
			(single) ? "single " : "multi", name);
		fflush(stdout);
	}
	if (single) {
		/* Single track load. */
		if ((tp = trackid(1, -1)) == NULL)
			return 0;
		capcheck(tp);
		if (load1(tp, ifp) == 0)
			return 0;
	} else {
		/* Multitrack load. */
		if (vflag)
			putchar('\n');
		capture = 0;
		iargc = 1;
		erase();
		while (fread(&i, sizeof(int), 1, ifp) == 1) {
			if (vflag)
				printf("\tTrack %d: ", i);
			if (load1(&track[i], ifp) == 0)
				return 0;
		}
	}
	status(0);
	if (fclose(ifp) == EOF)
		return error("cannot close input file");
	return 1;
}

/*
 * Load data from FILE to track 'tp'.
 * Return 1 for success, 0 for failure.
 */
int
load1(tp, ifp) register TRACK *tp; register FILE *ifp;
{
	register unsigned char *bp;
	register unsigned long size;
	register unsigned int usize;

	if (tp->t_buf != NULL)
		free(tp->t_buf);
	if (fread(tp, sizeof(TRACK), 1, ifp) != 1)
		return error("read error");
	if (vflag)
		printf("size=%ld\n", tp->t_size);
	bp = tp->t_buf = xalloc(tp->t_size);
	for (size = tp->t_size; size > 0L; ) {
		usize = (size > MAX_UINT) ? MAX_UINT : size;
		if (fread(bp, 1, usize, ifp) != usize)
			return error("read error");
		bp += usize;
		size -= usize;
	}
	return 1;
}

/*
 * Save one track or all tracks.
 * Interactive usage: s [ [ track ] dest ].
 */
int
save()
{
	register TRACK *tp;
	register FILE *ofp;
	int i, single;
	TRACK *ntp;
	char *name;
	unsigned long size;

	single = 0;
	if (iargc == 3) {
		++single;
		/* Save a single track. */
		if ((tp = trackid(1, 1)) == NULL)
			return;
		else if (isdigit(*iargv[2])) {

			/* Track to track. */
			if ((ntp = trackid(2, 0)) == NULL)
				return;
			capcheck(ntp);
			if (ntp->t_buf != NULL)
				free(ntp->t_buf);
			*ntp = *tp;
			ntp->t_buf = xalloc(tp->t_size);
			lmemcpy((char *)(ntp->t_buf), (char *)(tp->t_buf), tp->t_size);
			return;
		}
		name = iargv[2];
	} else if (iargc == 2) {
		name = iargv[1];
		filename = xalloc((unsigned long)strlen(name) + 1);
		strcpy(filename, name);
	} else {
		if (filename == NULL)
			return error("no default filename");
		name = filename;
	}

	if ((ofp = fopen(name, "wb")) == NULL)
		return error("cannot open output file \"%s\"", name);
	if (vflag) {
		printf("Save %strack file \"%s\": ",
			(single) ? "single" : "multi", name);
		fflush(stdout);
	}
	i = (single) ? ST_MAGIC : MT_MAGIC;
	if (fwrite(&i, sizeof(int), 1, ofp) != 1)
		return error("write error");
	size = 2L;
	if (single) {
		/* Single track to file. */
		if (save1(tp, ofp) == 0)
			return;
		size += sizeof(TRACK) + tp->t_size;
	} else {
		/* Multitrack to file. */
		for (i = 1; i <= NTRACKS; i++) {
			tp = &track[i];
			if (!inuse(tp))
				continue;
			if (fwrite(&i, sizeof(int), 1, ofp) != 1)
				return error("write error");
			if (save1(tp, ofp) == 0)
				return;
			size += sizeof(int) + sizeof(TRACK) + tp->t_size;
		}
		if (size == 2L)
			error("no work");
	}
	if (vflag)
		printf("size=%ld\n", size);
	if (fclose(ofp) == EOF)
		return error("cannot close output file \"%s\"", name);
}

/*
 * Save a track to a FILE.
 * Return success or failure.
 */
int
save1(tp, ofp) register TRACK *tp; register FILE *ofp;
{
	register unsigned char *bp;
	register unsigned int usize;
	unsigned long size;

	if (fwrite(tp, sizeof(TRACK), 1, ofp) != 1)
		return error("write error");
	for (bp = tp->t_buf, size = tp->t_size; size > 0L; ) {
		usize = (size > MAX_UINT) ? MAX_UINT : size;
		if (fwrite(bp, 1, usize, ofp) != usize)
			return error("write error");
		bp += usize;
		size -= usize;
	}
	return 1;
}

/* end of midifile.c */
