/* 
 * midimsg.c
 * (c) 1989 by Stephen A. Ness.
 * Print a MIDI message in readable form.
 */

#include <stdio.h>
#include "midimsg.h"

static char *pitch[12] = {
	"C", "C#", "D", "Eb", "E", "F",
	"F#", "G", "Ab", "A", "Bb", "B"
};

/*
 * Print a MIDI message.
 * Each message except System Exclusive is a single line.
 * Return a pointer past the end of the message.
 * This labels a NoteOn with velocity 0 as NoteOff,
 * a lie which makes the output more readable.
 */
unsigned char *
msg_print(bp) register unsigned char *bp;
{
	register unsigned char c;
	register char *s;
	register int n, type;

	c = *bp++;

	switch(Msg(c)) {

	case NoteOn:
		s = (*(bp+1) == 0) ? "NoteOff" : "NoteOn";
		type = 1;
		break;
	case NoteOff:		s = "NoteOff";		type = 1;	break;
	case PolyKeyPr:		s = "PolyKeyPr";	type = 1;	break;
	case CtrlChange:	s = "CtrlChange";	type = 2;	break;
	case ProgChange:	s = "ProgChange";	type = 3;	break;
	case ChanPr:		s = "ChanPr";		type = 3;	break;
	case PitchBend:		s = "PitchBend";	type = 4;	break;
	case SysEx:
		switch(c) {
		case SysEx:		s = "SysEx";		type = 5; break;
		case SongPosPtr:	s = "SongPosPtr";	type = 4; break;
		case SongSelect:	s = "SongSelect";	type = 3; break;
		case TuneRequest:	s = "TuneRequest";	type = 0; break;
		case EOX:		s = "EOX";		type = 0; break;
		case TimingClock:	s = "TimingClock";	type = 0; break;
		case Start:		s = "Start";		type = 0; break;
		case Continue:		s = "Continue";		type = 0; break;
		case Stop:		s = "Stop";		type = 0; break;
		case ActiveSensing:	s = "ActiveSensing";	type = 0; break;
		case SysReset:		s = "SysReset";		type = 0; break;
		default:	
			error("msg_print system exclusive botch %x", c);
			return bp;
		}
		break;

	default:
		error("msg_print botch %x", c);
		return bp;
	}

	printf("%-12s", s);			/* name */
	if (isSystem(c)) {
		if (type != 0 && type != 5)
			printf("      ");
	} else
		printf(" Ch %2d", Chan(c)+1);	/* channel */

	switch (type) {

	case 0:					/* Single byte commands */
		break;

	case 1:					/* NoteOff, NoteOn, PolyKeyPr */
		c = *bp++;			/* pitch */
		if (!isPitch(c))
			printf("<illegal pitch %d> ", c);
		else
			printf("%11s%d", pitch[note(c)], octave(c));
		printf("%12d", *bp++);		/* velocity */
		break;

	case 2:					/* CtrlChange */
		switch(c = *bp++) {		/* controller */
		case Modulation:	s = "Modulation";	break;
		case Volume:		s = "Volume";		break;
		case Pan:		s = "Pan";		break;
		case Expression:	s = "Expression";	break;
		case Sustain:		s = "Sustain";		break;
		case LocalCtrl:		s = "LocalCtrl";	break;
		case AllNotesOff:	s = "AllNotesOff";	break;
		case OmniModeOff:	s = "OmniModeOff";	break;
		case OmniModeOn:	s = "OmniModeOn";	break;
		case MonoModeOn:	s = "MonoModeOn";	break;
		case PolyModeOn:	s = "PolyModeOn";	break;
		default:		s = NULL;		break;
		}
		if (s != NULL)
			printf("%12s", s);
		else
			printf("%12d", c);
		printf("%12d", *bp++);		/* value */
		break;

	case 3:					/* ProgChange, ChanPr, SongSelect */
		printf("%12d", *bp++);		/* program or pressure */
		break;

	case 4:					/* PitchBend, SongPosPtr */
		n = *bp++;			/* lsb */
		n |= (*bp++) << 7;		/* msb */
		printf("%12d", n);
		break;

	case 5:					/* System exclusive */
		for (n = 0; (c = *bp++) != EOX; ) {
			printf("%02x ", c);
			if (++n % 16 == 0)
				printf("\n            ");
		}
		printf("\nEOX");
		break;
	}

	putchar('\n');
	return bp;
}

/* end of midimsg.c */
